""" 4.4 OAUTH 2 for devices

    Applications that run on devices with limited input capabilities 
    (such as game consoles, video cameras, and printers) can access a 
    Compliant API on behalf of a user, but the user must have separate 
    access to a computer or device with richer input capabilities. 
    The flow is as follows:

    +---------+                                  +---------------+
    |         |                                  |               |
    |  Your   |>--(A)---- Request Code --------->| Authorization |
    |  App    |                                  |     Server    |
    |         |<--(B)-- URL & Auth Code --------<|               |
    |         |                                  |               |
    |         |        (seperate device)         |               |
    |         |>--(C)- User login & consent ---->|               |
    |         |                                  |               |
    |         |>--(D)---- Poll server ---------->|               |
    |         |                                  |               |
    |         |<--(D)-- Token response ---------<|               |
    |         |                                  |               |
    +---------+                                  +---------------+

                    Figure 1: Device OAUTH2 Flow

    The flow illustrated in Figure 1 includes the following steps:

    (A)  Your application begins this flow with a request to a Service URL
         with a set of parameters. The response includes a device code,
         a user code, a URL, an expiration, and a suggested polling
         interval.

    (B) After receipt of this response, your application shows the user
        the URL and the user code, and instructs the user to open a
        browser, navigate to the URL, and enter the code.

    (C) The user switches to a device or computer with richer input
        capabilities, launches a browser, navigates to the URL
        specified on the limited-input device, logs in, and enters
        the code.

    (D) In the background, your application polls a Service endpoint
        for an access token This token will only be returned to your
        application after the user has logged in and approved the request.
"""

import logging
import functools
from flask import request
from flask import _request_ctx_stack as stack
from werkzeug import cached_property
import datetime
import json
from ..utility import create_response, decode_base64, json_serial

log = logging.getLogger('flask_oauth2-devices')

class OAuth2DevicesProvider(object):
    """
    Provide secure services for devices using OAuth2.
    
    There are two usage modes. One is
    binding the Flask app instance::

        app = Flask(__name__)
        oauth = OAuth2DevicesProvider(app)

    The second possibility is to bind the Flask app later::
        oauth = OAuth2DevicesProvider()
        def create_app():
            app = Flask(__name__)
            oauth.init_app(app)
            return app
    """    

    def __init__(self, app=None):
        self._before_request_funcs = []
        self._after_request_funcs = []
        self._invalid_response = None
        if app:
            self.init_app(app)

    def init_app(self, app):
        """
        This callback can be used to initialize an application for the
        oauth2 provider instance.
        """
        self.app = app
        app.extensions = getattr(app, 'extensions', {})
        app.extensions['oauth2devices.provider.oauth2devices'] = self

    @cached_property
    def error_uri(self):
        """The error page URI.
        """
        error_uri = self.app.config.get('OAUTH2_DEVICES_PROVIDER_ERROR_URI')
        if error_uri:
            return error_uri
        error_endpoint = self.app.config.get('OAUTH2_DEVICES_PROVIDER_ERROR_ENDPOINT')
        if error_endpoint:
            return url_for(error_endpoint)
        return '/oauth/errors'

    def clientgetter(self, f):
        """Register a function as the client getter.
        The function accepts one parameter `client_id`, and it returns
        a client object with at least these information:
            - client_id: A random string
            - client_secret: A random string
            - client_type: A string represents if it is `confidential`
            - redirect_uris: A list of redirect uris
        Implement the client getter:
            @oauth.clientgetter
            def get_client(client_id):
                client = get_client_model(client_id)
                # Client is an object
                return client
        """
        self._clientgetter = f
        return f

    def authcodesetter(self, f):
        """Register a function to save the auth code.

        The setter accepts five parameters, a least
            - code: our auth_code, if none we will generate one
            - client_id: the client we want to create a new auth_code for
            - user_id: the user we want to create a new auth_code for
        Implement the auth_code setter:
            @oauth.authcodesetter
            def save_auth_code(code, client_id, user_id, *args, **kwargs)
                auth_code_model.save_code(code, client, user_id)
        """
        self._authcodesetter = f
        return f

    def authcodegetter(self, f):
        """ Register a function as the client getter.

        The function accepts one parameter `code`, and it returns
        a code object.
        Implement the auth code getter::
            @oauth.authcodegetter
            def load_auth_code(code):
                code = get_code_model(code)
                # Code is an object
                return code
        """
        self._authcodegetter = f
        return f

    def code_handler(self, authorize_link, activate_link, expires_interval, polling_internal):
        """ Code handler decorator

        The device requests an auth_code as part of (A)
        
        For example, the client makes the following HTTP request using
        transport-only security (with extra line breaks for display purposes
        only):

            POST /oauth/device HTTP/1.1
            Host: server.example.com
            Content-Type: application/x-www-form-urlencoded

        The authorization server MUST authenticate the client. 
        """

        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                ctx = stack.top
                if ctx is not None and hasattr(ctx, 'request'):
                    request = ctx.request
                    if request.method != 'POST':
                        log.warn('Attempted a non-post on the code_handler')
                        return create_response({'Allow': 'POST'}, 'must use POST', 405)

                    app = self.getApp(request)

                    if app is None:
                        raise OAuth2Exception(
                            'Invalid application credentials',
                            type='unauthorized_client'
                        )

                    auth_code = self._authcodesetter(None, app.client_id, app.user_id)
                    return self.create_oauth2_code_response(auth_code, authorize_link, activate_link, expires_interval, polling_internal)

                return f(*args, **kwargs)
            return wrapper
        return decorator

    def authorize_handler(self):
        """ Authorize handler decorator

        The device uses the auth_code and device code it recieved from (A)
        and attempts to exchange it for an access token.

        For example, the client makes the following HTTP request using
        transport-layer security (with extra line breaks for display
        purposes only):

            POST /oauth/device/authorize HTTP/1.1
            Host: server.example.com
            Content-Type: application/x-www-form-urlencoded

            {
                "auth_code": "656ea891"
                "device_code: "c8fe9de9e6c5f80bc543c492aaa2fbaf2b081601"
            }
        """

        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                ctx = stack.top
                if ctx is not None and hasattr(ctx, 'request'):
                    request = ctx.request
                    if request.method != 'POST':
                        log.warn('Attempted a non-post on the code_handler')
                        return create_response({'Allow': 'POST'}, 'must use POST', 405)

                    data = request.values
                    auth_code = self._authcodegetter(data.get('auth_code'))

                    if auth_code is None:
                        raise OAuth2Exception(
                            'This token could not be found',
                            type='invalid_token'
                        )

                    if auth_code.expires is None and auth_code.expires < datetime.utcnow():
                        raise OAuth2Exception(
                            'Authorization code has expired',
                            type='invalid_token'
                        )

                    device_code = data.get('device_code')

                    if auth_code.is_active == 0:
                        raise OAuth2Exception(
                            'You have not authorized this device code yet',
                            type='not_authorized'
                        )

                    if auth_code.get_device_code() != device_code:
                        raise OAuth2Exception(
                            'Your user code does not match the device',
                            type='invalid_token'
                        )

                    access_token = auth_code.exchange_for_access_token(auth_code)
                    return self.create_oauth2_token_response(access_token)

                return f(*args, **kwargs)
            return wrapper
        return decorator

    def create_oauth2_code_response(self, auth_code, authorize_link=None, activate_link=None, expires_interval=0, polling_interval=0):
        """
        The authorization server issues an device code which the device will have
        prompt the user to authorize before following the activate link to
        exchange for a access token. The following parameters are added to the
        entity-body of the HTTP response with a 200 (OK) status code:

        device_code
            REQUIRED.  The device code generated on the fly for each device.

        user_code
            REQUIRED.  The auth code issued by the authorization server.

        authorize_link
            REQUIRED.  The link where auth code can be exchanged for access 
                       token.

        activate_link
            REQUIRED.  The link where auth code can be activated via user
                       consent flow.

        expires_in
            RECOMMENDED.  The lifetime in seconds of the access token.  For
            example, the value "3600" denotes that the access token will
            expire in one hour from the time the response was generated.
            If omitted, the authorization server SHOULD provide the
            expiration time via other means or document the default value.

        interval
            REQUIRED. The recommended polling interval.

        For example:

            HTTP/1.1 200 OK
            Content-Type: application/json;charset=UTF-8
            Cache-Control: no-store
            Pragma: no-cache

            {
                "device_code": "73de035b2a7bdcb2c092f4bdfe292898e0657a18",
                "user_code": "656e6075",
                "authorize_link": "https://api.example.com/oauth/device/authorize",
                "activate_link": "https://example.com/activate",
                "expires_in": 3600,
                "interval": 15
            }
        """
        response = create_response({
            'Content-Type': 'application/json',
            'Cache-Control': 'no-store',
            'Pragma': 'no-cache'}, json.dumps({
            'device_code' : auth_code.get_device_code(),
            'user_code ': auth_code.code,
            'authorize_link': authorize_link,
            'activate_link': activate_link,
            'expires_in': expires_interval,
            'interval': polling_interval}), 200)

        return response

    def create_oauth2_token_response(self, access_token):
        """
        The authorization server issues an access token and optional refresh
        token, and constructs the response by adding the following parameters
        to the entity-body of the HTTP response with a 200 (OK) status code:

        access_token
            REQUIRED.  The access token issued by the authorization server.

        token_type
            REQUIRED.  The type of the token issued as described in
            Section 7.1.  Value is case insensitive.

        expires_in
            RECOMMENDED.  The lifetime in seconds of the access token.  For
            example, the value "3600" denotes that the access token will
            expire in one hour from the time the response was generated.
            If omitted, the authorization server SHOULD provide the
            expiration time via other means or document the default value.

        refresh_token
            OPTIONAL.  The refresh token, which can be used to obtain new
            access tokens using the same authorization grant as described
            in Section 6.

        scope
            OPTIONAL, if identical to the scope requested by the client;
            otherwise, REQUIRED.  The scope of the access token as
            described by Section 3.3.

        The parameters are included in the entity-body of the HTTP response
        using the "application/json" media type as defined by [RFC4627].  The
        parameters are serialized into a JavaScript Object Notation (JSON)
        structure by adding each parameter at the highest structure level.
        Parameter names and string values are included as JSON strings.
        Numerical values are included as JSON numbers.  The order of
        parameters does not matter and can vary.

        The authorization server MUST include the HTTP "Cache-Control"
        response header field [RFC2616] with a value of "no-store" in any
        response containing tokens, credentials, or other sensitive
        information, as well as the "Pragma" response header field [RFC2616]
        with a value of "no-cache".

        For example:

            HTTP/1.1 200 OK
            Content-Type: application/json;charset=UTF-8
            Cache-Control: no-store
            Pragma: no-cache

            {
                "access_token":"2YotnFZFEjr1zCsicMWpAA",
                "token_type":"example",
                "scope":"public private",
                "expires_in":3600,
                "refresh_token":"tGzv3JOkF0XG5Qx2TlKWIA"
            }

        The client MUST ignore unrecognized value names in the response.  The
        sizes of tokens and other values received from the authorization
        server are left undefined.  The client should avoid making
        assumptions about value sizes.  The authorization server SHOULD
        document the size of any value it issues.

        http://tools.ietf.org/html/rfc6749#section-5.1
        """
        response = create_response({
            'Content-Type': 'application/json',
            'Cache-Control': 'no-store',
            'Pragma': 'no-cache'}, json.dumps({
            'access_token' : access_token.access_token,
            'token_type ': access_token.token_type,
            'scope': access_token.scopes,
            'expires_in': json.dumps(access_token.expires, default=json_serial).replace("\"", ""),
            'refresh_token': None}), 200)

        return response

    def getApp(self, request):
        # http://tools.ietf.org/html/rfc2617#section-2

        client_id = None
        client_secret = None

        if "Authorization" in request.headers:
            auth_header = request.headers['Authorization']

            if "basic" in auth_header:
                auth = decode_base64(auth_header[6:]).split(':')
                client_id = auth[0]
                client_secret = auth[1]

        if client_id is None:
            raise OAuth2Exception(
                'A valid client ID must be provided along with request made',
                type='invalid_client'
            )

        app = self._clientgetter(client_id)

        if app is None:
            raise OAuth2Exception(
                'A valid client ID must be provided along with request made',
                type='invalid_client'
            )

        if client_secret is not None and client_secret == app.client_secret:
            return app

        raise OAuth2Exception(
            'A valid client secret must be provided along with request made',
            type='invalid_secret'
        )


class OAuth2Exception(RuntimeError):
    def __init__(self, message, type=None, data=None):
        self.message = message
        self.type = type
        self.data = data

    def __str__(self):
        return self.message.encode('utf-8')



import base64
from datetime import datetime
from flask import request, Response
from oauthlib.common import to_unicode, bytes_type

def to_bytes(text, encoding='utf-8'):
    """Make sure text is bytes type."""
    if not text:
        return text
    if not isinstance(text, bytes_type):
        text = text.encode(encoding)
    return text

def decode_base64(text, encoding='utf-8'):
    """Decode base64 string."""
    text = to_bytes(text, encoding)
    return to_unicode(base64.b64decode(text), encoding)

def create_response(headers, body, status):
    """Create response class for Flask."""
    response = Response(body or '')
    for k, v in headers.items():
        response.headers[str(k)] = v

    response.status_code = status
    return response

def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial



import OpenSSL
import hmac
import hashlib
from binascii import hexlify
from datetime import datetime, timedelta
from flask import Flask, abort, render_template, make_response, request
from flask_sqlalchemy import SQLAlchemy
from flask.ext.wtf import Form
from wtforms import StringField, SelectField
from wtforms.validators import DataRequired
from oauth2devices import OAuth2DevicesProvider, OAuth2Exception

from forms import ActivateForm

app = Flask(__name__)
app.config.update(
    WTF_CSRF_ENABLED = True,
    SECRET_KEY = 'our-big-bad-key'
)
app.config.update({
    'SQLALCHEMY_DATABASE_URI': 'sqlite:///db.sqlite'
})
db = SQLAlchemy(app)
oauth = OAuth2DevicesProvider(app)

AUTH_EXPIRATION_TIME = 3600
OUR_KEY = 'our-big-bad-key'

@app.route('/oauth/device', methods=['POST'])
@oauth.code_handler("https://api.example.com/oauth/device/authorize", "https://example.com/activate", 600, 600)
def code():
    return None

@app.route('/oauth/device/authorize', methods=['POST'])
@oauth.authorize_handler()
def authorize():
    return None

@app.route('/activate', methods=['GET', 'POST'])
def activate_view():

    form = ActivateForm()
    if form.validate_on_submit():
        if request.method == "POST":

            user_code = load_auth_code(request.values.get('user_code'))

            if user_code is None or user_code.expires < datetime.utcnow():
                return render_template('app_auth_error.html')

            return redirect("/oauth/authorization/accept?user_code="+str(user_code.code))

    resp = make_response(render_template('user_code_activate.html', form=form))
    resp.headers.extend({'X-Frame-Options': 'DENY'})
    return resp

@app.route('/oauth/authorization/accept', methods=['GET', 'POST'])
def authorization_accept_view():

    user_code = load_auth_code(request.values.get('user_code'))

    all_scopes = ['private']

    # public is our default scope in this case
    if request.values.get('scopes') is None:
        scopes = ['public']
    else:
        scopes = request.values.get('scopes').split()

    non_scopes = [scope for scope in all_scopes if scope not in scopes]

    resp = make_response(render_template('access_token_authorize.html',
                                         app_id=user_code.client_id,
                                         client_id=user_code.client_id,
                                         user_code=user_code.code,
                                         scopes=scopes,
                                         non_scopes=non_scopes))
    resp.headers.extend({'X-Frame-Options': 'DENY'})
    return resp

@app.route('/confirmed', methods=['POST'])
def confirmed_view():

    # just an extra check in case we didn't block GET in the decorator
    if request.method != "POST":
        resp = make_response("non-POST on access token", 405)
        resp.headers.extend({'Allow': 'POST'})
        return resp

    client_id = request.values.get('client_id')

    if client_id is None:
        return make_response("missing client_id", 500)

    # we can load our app by client_id here 
    # and throw a 500 if we have a problem

    user_code = load_auth_code(request.values.get('user_code'))

    if user_code is None:
        return make_response("auth code must be sent", 400)

    user_code.is_active = 1
    db.session.commit()

    resp = make_response(render_template('app_auth_confirm.html', client_id=user_code.client_id))
    resp.headers.extend({'X-Frame-Options': 'DENY'})
    return resp

class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(40), unique=True)

class Client(db.Model):
    client_id = db.Column(db.String(40), primary_key=True)
    client_secret = db.Column(db.String(55), nullable=False)

    user_id = db.Column(db.ForeignKey('user.id'))
    user = db.relationship('User')

    _redirect_uris = db.Column(db.Text)
    _default_scopes = db.Column(db.Text)

    @property
    def client_type(self):
        return 'public'

    @property
    def redirect_uris(self):
        if self._redirect_uris:
            return self._redirect_uris.split()
        return []

    @property
    def default_redirect_uri(self):
        return self.redirect_uris[0]

    @property
    def default_scopes(self):
        if self._default_scopes:
            return self._default_scopes.split()
        return []

class Token(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    client_id = db.Column(
        db.String(40), db.ForeignKey('client.client_id'),
        nullable=False,
    )
    client = db.relationship('Client')

    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id')
    )
    user = db.relationship('User')

    # currently only bearer is supported
    token_type = db.Column(db.String(40))

    access_token = db.Column(db.String(255))
    refresh_token = db.Column(db.String(255))
    expires = db.Column(db.DateTime)
    created = db.Column(db.DateTime)
    _scopes = db.Column(db.Text)

    @property
    def scopes(self):
        if self._scopes:
            return self._scopes.split()
        return []

    def create_access_token(self, client_id, user_id, scope, token_type):

        expires_in = AUTH_EXPIRATION_TIME
        expires = datetime.utcnow() + timedelta(seconds=expires_in)
        created = datetime.utcnow()

        tok = Token(
            client_id=client_id,
            user_id=user_id,
            access_token=None,
            refresh_token=None,
            token_type=token_type,
            _scopes = ("public private" if scope is None else ' '.join(scope)),
            expires=expires,
            created=created,
        )

        if tok.access_token is None:
            tok.access_token = tok._generate_token()

        db.session.add(tok)
        db.session.commit()
        return tok

    def refresh(self, token):

        tok = Token(
            client_id=self.client_id,
            user_id=self.user_id,
            access_token=self.access_token,
            refresh_token=None,
            token_type=token_type,
            _scopes = ("public private" if scope is None else ' '.join(scope)),
            expires=expires,
            created=created,
        )

        if tok.refresh_token is None:
            tok.refresh_token = tok._generate_refresh_token()

        db.session.add(tok)
        db.session.commit()
        return tok

    def _generate_token(self):
        return hashlib.sha1("app:" + str(self.client_id) + ":user:" + str(self.user_id) + str(hexlify(OpenSSL.rand.bytes(10)))).hexdigest()

    def _generate_refresh_token(self):
        return hashlib.sha1("app:" + str(self.client_id) + ":user:" + str(self.user_id) + ":access_token:" + str(self.id)).hexdigest()

    def contains_scope(scope):
        return scope in self.scope.split(' ')

class Code(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    client_id = db.Column(
        db.String(40), db.ForeignKey('client.client_id'),
        nullable=False,
    )
    client = db.relationship('Client')

    user_id = db.Column(
        db.Integer, db.ForeignKey('user.id')
    )
    user = db.relationship('User')

    code = db.Column(db.String(40), unique=True)
    _scopes = db.Column(db.Text)
    expires = db.Column(db.DateTime)
    created = db.Column(db.DateTime)
    is_active = db.Column(db.Integer)

    @property
    def scopes(self):
        if self._scopes:
            return self._scopes.split()
        return []

    def generate_new_code(self, client_id):
        return hashlib.sha1("secret:" + client_id + ":req:" + str(hexlify(OpenSSL.rand.bytes(10)))).hexdigest()

    def get_device_code(self):
        return hmac.new(OUR_KEY, "secret:"+str(self.id), hashlib.sha1).hexdigest()

    def exchange_for_access_token(self, app):
        return Token().create_access_token(app.client_id, app.user_id, app.scopes, "grant_auth_code")

def current_user():
    if 'id' in session:
        uid = session['id']
        return User.query.get(uid)
    if 'id' in request.args:
        return User.query.get(request.args.get('id'))
    return None

@oauth.clientgetter
def load_client(client_id):
    return Client.query.filter_by(client_id=client_id).first()

@oauth.authcodesetter
def save_auth_code(code, client_id, user_id, *args, **kwargs):
    codes = Code.query.filter_by(
        client_id=client_id,
        user_id=user_id
    )

    # make sure that every client has only one code connected to a user
    for c in codes:
        db.session.delete(c)

    expires_in = (AUTH_EXPIRATION_TIME if code is None else code.pop('expires_in'))
    expires = datetime.utcnow() + timedelta(seconds=expires_in)
    created = datetime.utcnow()

    cod = Code(
        client_id=client_id,
        user_id=user_id,
        code = (None if code is None else code['code']),
        _scopes = ('public private' if code is None else code['scope']),
        expires=expires,
        created=created,
        is_active=0
    )

    if cod.code is None:
        cod.code = cod.generate_new_code(cod.client_id)[:8]

    db.session.add(cod)
    db.session.commit()
    return cod

@oauth.authcodegetter
def load_auth_code(code):
    return Code.query.filter_by(code=code).first()

if __name__ == "__main__":
    db.create_all()
    app.run(debug=True)



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup
import os

long_description = 'OAuth2 for devices flow implementation for Flask'
if os.path.exists('README.rst'):
    long_description = open('README.rst').read()

setup(
    name='flask-oauth2-devices',
    version='0.0.1',
    description='OAuth2 for devices flow implementation for Flask',
    author='Joe Cabrera',
    author_email='jcabrera@eminorlabs.com',
    url='https://github.com/greedo/flask-oauth2-devices',
    license='MIT License',
    keywords='Flask,    ',
    packages=[
        'devices',
        'devices.provider'],
    install_requires=[
        'pytest',
        'pep8',
        'Flask',
        'Flask-SQLAlchemy',
        'pyOpenSSL',
        'six'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    long_description=long_description
)



import six
import base64
from datetime import datetime
from flask import request, Response

if six.PY3:
    unicode_type = str
    bytes_type = bytes
else:
    unicode_type = unicode
    bytes_type = str

def to_unicode(data):
    """Convert utf-8 to unicode."""
    if isinstance(data, bytes_type):
        return unicode_type(data, 'UTF-8')

    if hasattr(data, '__iter__'):
        try:
            dict(data)
        except TypeError:
            pass
        except ValueError:
            return (to_unicode(i, 'UTF-8') for i in data)
        else:
            if hasattr(data, 'items'):
                data = data.items()
            return dict(((to_unicode(k, 'UTF-8'), to_unicode(v, 'UTF-8')) for k, v in data))

    return data

def to_bytes(text):
    """Make sure text is bytes type."""
    if not text:
        return text
    if not isinstance(text, bytes_type):
        text = text.encode('utf-8')
    return text

def decode_base64(text):
    """Decode base64 string from utf-8."""
    text = to_bytes(text)
    return to_unicode(base64.b64decode(text))

def create_response(headers, body, status):
    """Create response object for Flask."""
    response = Response(body or '')
    for k, v in headers.items():
        response.headers[str(k)] = v

    response.status_code = status
    return response

def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
